
from PySide.QtCore import *
from PySide.QtGui import *

from .formatter import Formatter
from .server import Server
from .browser import Browser
from .dialogs import Info, Settings
from .ui.main import Ui_main
from .__version__ import version


class Qygmy(QMainWindow):

    def __init__(self):
        super().__init__()
        QTextCodec.setCodecForTr(QTextCodec.codecForName('UTF-8'))
        self.settings = Settings(self)
        self.fmt = Formatter(self.settings)
        self.srv = Server(self)
        self.info = Info(self)
        self.browser = Browser(self)
        self.timer = QTimer(self)
        self.setup_ui()
        self.timer.start(int(self.settings['gui']['interval']))

    def connect_mpd(self):
        self.srv.connect_mpd(
            self.settings['connection']['host'],
            int(self.settings['connection']['port']),
            self.settings['connection']['password'])

    def create_eventloop_timer(self, slot):
        t = QTimer(self)
        t.setInterval(0)
        t.setSingleShot(True)
        t.timeout.connect(slot)
        return t

    def setup_ui(self):
        self.ui = Ui_main()
        self.ui.setupUi(self)
        self.ui.current_song.setText(self.fmt.current_song('disconnect', {}))
        self.setup_icons()
        self.setup_widgets()

        self.busy_stack = 0
        self.busy_stack_real = 0
        self.notbusy_timer = self.create_eventloop_timer(self.not_busy2)
        self.scroll_timer = self.create_eventloop_timer(self.scroll_to_current2)
        self.setup_signals()

        try:
            if 'main_geometry' in self.settings['guistate']:
                self.restoreGeometry(QByteArray.fromBase64(self.settings['guistate']['main_geometry']))
        except:
            pass

    def setup_icons(self):
        self.setWindowIcon(QIcon.fromTheme('applications-multimedia'))
        self.ui.settings_icon = QIcon.fromTheme('configure')
        self.ui.settings_icon_updating = QIcon.fromTheme('view-refresh')
        for e, icon in (
            ('previous', 'media-skip-backward'),
            ('play', 'media-playback-start'),
            ('pause', 'media-playback-pause'),
            ('stop', 'media-playback-stop'),
            ('next', 'media-skip-forward'),
            ('volume', 'audio-volume-high'),
            ('add', 'list-add'),
            ('remove', 'list-remove'),
            ('clear', 'edit-clear-list'),
            ('repeat', 'media-playlist-repeat'),
            ('shuffle', 'media-playlist-shuffle'),
            ('single', 'go-last'),
            ('consume', 'edit-delete-shred'),
            ('highprio', 'flag-red'),
            ('save', 'document-save'),
            ('details', 'dialog-information'),
            ('settings', 'configure'),
            ('updatedb', 'view-refresh'),
            ('about', 'applications-multimedia'),
            ('quit', 'application-exit'),
            ('louder', 'audio-volume-high'),
            ('quieter', 'audio-volume-low'),
            ('playback_menu', 'media-playback-start'),
            ('playlist_menu', 'text-plain'),
        ):
            getattr(self.ui, e).setIcon(QIcon.fromTheme(icon))

    def setup_widgets(self):
        self.ui.playback_toolbar.insertWidget(self.ui.volume, self.ui.progressbar)
        self.ui.queue_toolbar.insertWidget(self.ui.settings, self.ui.status)
        self.ui.queue_toolbar.insertSeparator(self.ui.settings)
        self.ui.queue.setup(self.srv.queue)

        self.ui.wa1 = QWidgetAction(self)
        self.ui.wa1.setDefaultWidget(self.ui.volume_label)
        self.ui.wa2 = QWidgetAction(self)
        self.ui.wa2.setDefaultWidget(self.ui.volume_slider)
        vm = QMenu(self)
        vm.addAction(self.ui.wa1)
        vm.addAction(self.ui.wa2)
        self.ui.volume_popup = vm
        vb = self.ui.playback_toolbar.widgetForAction(self.ui.volume)
        vb.setMenu(vm)
        vb.setPopupMode(QToolButton.InstantPopup)

    def setup_signals(self):
        self.timer.timeout.connect(self.srv.update_state)
        self.ui.play.triggered.connect(self.srv.play)
        self.ui.pause.triggered.connect(self.srv.pause)
        self.ui.stop.triggered.connect(self.srv.stop)
        self.ui.next.triggered.connect(self.srv.next)
        self.ui.previous.triggered.connect(self.srv.previous)
        self.ui.progressbar.clicked.connect(self.srv.times.send)
        self.srv.times.changed.connect(self.update_progressbar)
        self.srv.state.changed.connect(self.update_progressbar)
        self.srv.state.changed.connect(self.update_current_song)
        self.srv.current_song.changed.connect(self.update_progressbar)
        self.srv.current_song.changed.connect(self.update_current_song)
        self.srv.current_song.changed2.connect(self.scroll_to_current)
        self.srv.queue.current.changed.connect(self.update_status)
        self.srv.volume.changed.connect(self.ui.volume_slider.setValue)
        self.srv.volume.changed.connect(lambda v: self.ui.volume_label.setText(str(v)))
        self.ui.volume_slider.sliderMoved.connect(self.srv.volume.send)
        self.ui.louder.triggered.connect(lambda v=self.srv.volume: v.send(min(v.value+1, 100)))
        self.ui.quieter.triggered.connect(lambda v=self.srv.volume: v.send(max(v.value-1, 0)))
        for v in ('repeat', 'shuffle', 'single', 'consume'):
            act = getattr(self.ui, v)
            var = getattr(self.srv, v)
            def slot(checked, act=act, var=var):
                act.setChecked(var.value)
                var.send(checked)
            act.triggered[bool].connect(slot)
            var.changed.connect(act.setChecked)
        self.ui.add.triggered.connect(self.browser.show)
        self.ui.add.triggered.connect(self.browser.activateWindow)
        self.ui.remove.triggered.connect(self.ui.queue.remove_selected)
        self.ui.clear.triggered.connect(self.srv.clear)
        self.ui.randomize.triggered.connect(self.srv.randomize_queue)
        self.ui.reverse.triggered.connect(self.srv.queue.reverse)
        self.ui.updatedb.triggered.connect(self.srv.updatedb)
        self.ui.connect.triggered.connect(self.connect_mpd)
        self.ui.disconnect.triggered.connect(self.srv.disconnect_mpd)
        self.ui.settings.triggered.connect(self.settings.exec_)
        self.ui.quit.triggered.connect(self.close)
        self.ui.highprio.triggered.connect(lambda: self.ui.queue.set_priority(1))
        self.ui.normprio.triggered.connect(lambda: self.ui.queue.set_priority(0))

    def error(self, message):
        QMessageBox.critical(self, self.tr('Error'), self.tr(message))

    def update_progressbar(self, *_):
        s = self.srv.state.value
        elapsed, total = self.srv.times.value
        text = self.fmt.progressbar(s, elapsed, total, self.srv.current_song.value)
        if elapsed > total:
            total = elapsed
        if (elapsed, total) == (0, 0):
            total = 1
        self.ui.progressbar.setRange(0, total)
        self.ui.progressbar.setValue(elapsed)
        self.ui.progressbar.setFormat(text)

    def update_current_song(self, *_):
        s, c = self.srv.state.value, self.srv.current_song.value
        self.ui.current_song.setText(self.fmt.current_song(s, c))
        self.ui.current_song.setToolTip(self.fmt.current_song_tooltip(s, c))
        self.setWindowTitle(self.fmt.window_title(s, c))

    def update_status(self, *_):
        self.ui.status.setText(self.fmt.status(self.srv.state.value,
                self.srv.queue.total_length, len(self.srv.queue)))

    def scroll_to_current(self, new_song, old_song):
        pnew, pold = new_song.get('file', ''), old_song.get('file', '')
        pos = int(new_song.get('pos', '-1'))
        if (pos < 0 or pnew == pold or not pnew or
                self.srv.state.value not in ('play', 'pause') or
                self.settings['gui']['autoscroll'] != '1'):
            return
        self.scroll_pos = pos
        self.scroll_timer.start()

    def scroll_to_current2(self):
        pos = self.scroll_pos
        index = self.srv.queue.index(pos, 0)
        if not self.ui.queue.rect().contains(self.ui.queue.visualRect(index)):
            self.ui.queue.scrollTo(index, QAbstractItemView.PositionAtCenter)

    def contextMenuEvent(self, e):
        e.accept()
        self.ui.highprio.setEnabled(self.ui.queue.can_set_priority(1))
        self.ui.normprio.setEnabled(self.ui.queue.can_set_priority(0))
        self.ui.details.setEnabled(self.ui.queue.details() is not None)
        self.ui.context_menu.popup(e.globalPos())

    def closeEvent(self, e):
        self.srv.disconnect_mpd()
        self.settings['guistate']['main_geometry'] = str(self.saveGeometry().toBase64())
        self.browser.close()
        self.settings.save()
        super().closeEvent(e)

    @Slot(str)
    def on_state_changed(self, state):
        c = state != 'disconnect'
        self.ui.connect.setVisible(not c)
        self.ui.disconnect.setVisible(c)
        self.ui.play.setVisible(state != 'play')
        self.ui.pause.setVisible(state == 'play')
        for e in ('previous', 'play', 'pause', 'stop', 'next', 'volume',
                'add', 'clear', 'repeat', 'shuffle', 'single', 'consume',
                'updatedb', 'save', 'randomize', 'details', 'statistics',
                'louder', 'quieter', 'reverse',
                'playback_menu', 'volume_menu', 'playlist_menu'):
            getattr(self.ui, e).setEnabled(c)
        self.on_queue_selection_changed()

    @Slot(bool)
    def on_updating_db_changed(self, updating):
        if updating:
            self.ui.settings.setIcon(self.ui.settings_icon_updating)
        else:
            self.ui.settings.setIcon(self.ui.settings_icon)

    @Slot()
    def on_queue_selection_changed(self):
        self.ui.remove.setEnabled(self.srv.state.value != 'disconnect' and
                self.ui.queue.can_remove())

    @Slot()
    def on_details_triggered(self):
        d = self.ui.queue.details()
        if d is not None:
            self.info.exec_('details', d, 'Details')

    @Slot()
    def on_statistics_triggered(self):
        self.info.exec_('statistics', self.srv.statistics(), 'MPD statistics')

    @Slot()
    def on_save_triggered(self):
        name = None
        while True:
            name, ok = QInputDialog.getText(self,
                    self.tr('Save current playlist'),
                    self.tr('Playlist name:'),
                    text=name if name else self.srv.playlists.NEW_PLAYLIST_NAME)
            if not ok:
                break
            if not name:
                self.error('You have to provide a name.')
            elif self.srv.playlists.save_queue(name) is False:
                ans = QMessageBox.question(self, self.tr('Error'),
                        self.tr('Playlist with such name already exists. '
                        'Do you want to replace it?'),
                        QMessageBox.Yes | QMessageBox.No)
                if ans == QMessageBox.Yes:
                    self.srv.playlists.save_queue(name, True)
                    break
            else: break

    @Slot()
    def on_aboutqt_triggered(self):
        QMessageBox.aboutQt(self)

    @Slot()
    def on_about_triggered(self):
        QMessageBox.about(
            self, self.tr('About Qygmy'),
            self.tr(
                '<h2>Qygmy</h2>'
                '<p>version {version}</p>'
                '<p>Simple MPD client written in Python and Qt/PySide.</p>'
                '<p><a href="https://github.com/tsufeki/qygmy">https://github.com/tsufeki/qygmy</a></p>'
            ).format(version=version))

    def busy(self):
        self.busy_stack_real += 1
        self.busy_stack += 1
        QApplication.setOverrideCursor(Qt.BusyCursor)

    def not_busy(self):
        self.busy_stack -= 1
        self.notbusy_timer.start()

    def not_busy2(self):
        while self.busy_stack_real > self.busy_stack:
            self.busy_stack_real -= 1
            QApplication.restoreOverrideCursor()


def main(translations_path=None):
    import sys
    import os

    if translations_path is None:
        translations_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'translations')

    app = QApplication(sys.argv)
    locale = QLocale.system().name()

    qt_tr = QTranslator()
    if qt_tr.load('qt_' + locale, QLibraryInfo.location(QLibraryInfo.TranslationsPath)):
        app.installTranslator(qt_tr)

    qygmy_tr = QTranslator()
    if qygmy_tr.load('qygmy_' + locale, translations_path):
        app.installTranslator(qygmy_tr)

    m = Qygmy()
    m.show()
    m.connect_mpd()
    # XXX: TODO: FIXME:
    os._exit(app.exec_())
    #sys.exit(app.exec_())

if __name__ == '__main__':
    main()





import os
import configparser

from PySide.QtCore import *
from PySide.QtGui import *

from .ui.infodialog import Ui_infodialog
from .ui.settings import Ui_settings

import logging
logger = logging.getLogger('qygmy')


class Info(QDialog):
    def __init__(self, main):
        super().__init__(main)
        self.main = main
        self.ui = Ui_infodialog()
        self.ui.setupUi(self)
        self.ui.icon_label.setPixmap(QIcon.fromTheme('dialog-information').pixmap(64, 64))

    def exec_(self, name, info, title):
        self.setWindowTitle(title)
        while self.ui.layout.count() > 0:
            w = self.ui.layout.takeAt(0)
            w.widget().hide()
            w.widget().deleteLater()
        d = self.main.fmt.info_dialog(name, info)
        for a, b in d:
            if b is None:
                self.ui.layout.addRow(QLabel(a))
            else:
                self.ui.layout.addRow(a, QLabel(b))
        self.adjustSize()
        super().exec_()


class Settings(QDialog):

    PATH = os.path.expanduser(os.path.join(os.environ.get('XDG_CONFIG_HOME', '~/.config'), 'qygmy'))
    FILENAME = 'qygmy.conf'

    def __init__(self, main):
        super().__init__(main)
        self.main = main
        self.ui = Ui_settings()
        self.ui.setupUi(self)
        self.retranslate()

        self.conf = configparser.ConfigParser(interpolation=None)
        self.conf.read_dict(self.defaults, '<defaults>')
        self.conf.read_dict(self.environ_conf(), '<MPD_HOST>')
        self.conf.read(os.path.join(self.PATH, self.FILENAME), 'utf-8')
        self.validate(self)
        self.conf_to_ui()

    def retranslate(self):
        self.defaults = {
            'connection': {
                'host': 'localhost',
                'port': '6600',
                'password': '',
            },
            'format': {
                'window_title': self.tr(
                    'Qygmy'
                    '$if(%playing%, / '
                    '$if(%artist%,%artist% \u2014 )'
                    '$if2(%title%,%filename%))'
                ),
                'progressbar': self.tr(
                    '$if3('
                        '%playing%%paused%,'
                            '$time(%elapsed%)'
                            '$if($and(%total%,$gt(%total%,0)), / $time(%total%)),'
                        '%stopped%,Stopped,'
                        '%connected%,Connected,'
                        'Disconnected)'
                ),
                'current_song': self.tr(
                    '<span style="font-size: large; font-weight: bold">'
                        '$if2(%title%,%filename%)</span><br>'
                    '%artist%$if(%album%, \u2014 %album%)'
                ),
                'playlist_item': self.tr(
                    '$if(%artist%,%artist% \u2014 )'
                    '$if2(%title%,%filename%)'
                ),
            },
            'gui': {
                'autoscroll': '1',
                'interval': '500',
            },
            'guistate': {},
        }
        self.ui.retranslateUi(self)

    def environ_conf(self):
        if 'MPD_HOST' in os.environ:
            env = os.environ['MPD_HOST']
            c = {}
            if '@' in env:
                c['password'], env = env.rsplit('@', 1)
            if ':' in env:
                env, c['port'] = env.rsplit(':', 1)
            if env:
                c['host'] = env
            return {'connection': c}
        return {}

    def __getitem__(self, section):
        return self.conf[section]

    def __contains__(self, section):
        return section in self.conf

    def exec_(self):
        self.conf_to_ui()
        super().exec_()

    @Slot(QAbstractButton)
    def on_buttonbox_clicked(self, button):
        sb = self.ui.buttonbox.standardButton(button)
        if sb == QDialogButtonBox.Ok:
            self.ui_to_conf()
            self.save()
            self.conf_to_ui()
        elif sb == QDialogButtonBox.Cancel:
            self.conf_to_ui()
        elif sb == QDialogButtonBox.Apply:
            self.ui_to_conf()
            self.save()
            self.conf_to_ui()
        elif sb == QDialogButtonBox.RestoreDefaults:
            self.conf_to_ui(self.defaults)

    def conf_to_ui(self, conf=None):
        if conf is None:
            conf = self.conf
        c, f, g = conf['connection'], conf['format'], conf['gui']
        self.ui.host.setText(c['host'])
        self.ui.port.setText(c['port'])
        self.ui.password.setText(c['password'])
        self.ui.window_title.setPlainText(f['window_title'])
        self.ui.progressbar.setPlainText(f['progressbar'])
        self.ui.current_song.setPlainText(f['current_song'])
        self.ui.playlist_item.setPlainText(f['playlist_item'])
        self.ui.autoscroll.setChecked(g['autoscroll'] == '1')
        self.ui.interval.setValue(int(g['interval']))

    def ui_to_dict(self):
        return {
            'connection': {
                'host': self.ui.host.text(),
                'port': self.ui.port.text(),
                'password': self.ui.password.text(),
            },
            'format': {
                'window_title': self.ui.window_title.toPlainText(),
                'progressbar': self.ui.progressbar.toPlainText(),
                'current_song': self.ui.current_song.toPlainText(),
                'playlist_item': self.ui.playlist_item.toPlainText(),
            },
            'gui': {
                'autoscroll': '1' if self.ui.autoscroll.isChecked() else '0',
                'interval': str(self.ui.interval.value()),
            },
        }

    def ui_to_conf(self):
        d = self.ui_to_dict()
        self.validate(d)
        changed = set()
        for sect in ('connection', 'format', 'gui'):
            for k in d[sect]:
                if d[sect][k] != self.conf[sect][k]:
                    changed.add(k)
                    self.conf[sect][k] = d[sect][k]
        if not {'host', 'port', 'password'}.isdisjoint(changed) and self.main.srv.state != 'disconnect':
            self.main.srv.disconnect_mpd()
            self.main.connect_mpd()
        if not {'window_title', 'progressbar', 'current_song', 'playlist_item'}.isdisjoint(changed):
            self.main.fmt.clear_cache()
        if not {'window_title', 'current_song'}.isdisjoint(changed):
            self.main.update_current_song()
        if 'progressbar' in changed:
            self.main.update_progressbar()
        if 'playlist_item' in changed:
            self.main.srv.queue.refresh_format()
            self.main.srv.database.refresh_format()
            self.main.srv.playlists.refresh_format()
            self.main.srv.search.refresh_format()
        if 'interval' in changed:
            self.main.timer.stop()
            self.main.timer.start(int(self.conf['gui']['interval']))

    def save(self):
        try:
            os.makedirs(self.PATH, exist_ok=True)
            with open(os.path.join(self.PATH, self.FILENAME), 'w', encoding='utf-8') as f:
                self.conf.write(f)
        except OSError as e:
            logger.error('{}: {}'.format(e.__class__.__name__, e))

    @classmethod
    def validate_int(cls, string, imin, imax, default):
        string = string.strip()
        try:
            i = int(string)
            if imin <= i < imax:
                return string
        except ValueError:
            pass
        return default

    def validate(self, dct):
        c, g = dct['connection'], dct['gui']
        c['host'] = c['host'].strip()
        c['port'] = self.validate_int(c['port'], 1, 2**16, self.defaults['connection']['port'])
        g['autoscroll'] = '1' if g['autoscroll'] == '1' else '0'
        g['interval'] = self.validate_int(g['interval'], 100, 3600000, self.defaults['gui']['interval'])





import re


# Conditionals

def lazy_if(cond, yes, no=lambda:''): return yes() if cond() else no()

def lazy_if2(*args):
    """Return first non-empty argument."""
    for a in args:
        val = a()
        if val:
            return val
    return ''

def lazy_if3(*args):
    """Shorthand for a chain of `if`s.

    `$if3(c_0, v_0, c_1, v_1, c_2, v_2, ...)` - return first `v_i` for which
    `c_i` is not empty.
    """
    if len(args) == 0:
        return ''
    if len(args) == 1:
        return args[0]()
    return lazy_if(args[0], args[1], lambda: lazy_if3(*args[2:]))


# String operations

def f_lower(text): return text.lower()
def f_upper(text): return text.upper()
def f_left(text, num): return text[:int(num)]
def f_right(text, num): return text[-int(num):]
def f_num(num, length): return '{{:0{:d}d}}'.format(int(length)).format(int(num))
def f_replace(text, search, replace): return text.replace(search, replace)
def f_in(x, y): return '1' if x in y else ''
def f_trim(text, chars=None): return text.strip(chars)
def f_len(text): return str(len(text))

def f_eq(x, y): return '1' if x == y else ''
def f_ne(x, y): return '1' if x != y else ''


# Regular expressions

def f_rsearch(text, pattern):
    m = re.search(pattern, text)
    if m is None:
        return ''
    if m.lastindex is None:
        return m.group(0)
    return m.group(1)

def f_rreplace(text, pattern, replace):
    return re.sub(pattern, replace, text)


# Integer arithmetic

def f_add(*args): return str(sum(int(i) for i in args))
def f_sub(x, y): return str(int(x) - int(y))
def f_div(x, y): return str(int(x) // int(y))
def f_mod(x, y): return str(int(x) % int(y))

def f_mul(*args):
    r = 1
    for i in args:
        r *= int(i)
    return str(r)

def f_max(*args): return str(max(int(i) for i in args))
def f_min(*args): return str(min(int(i) for i in args))

def f_lt(x, y): return '1' if int(x) < int(y) else ''
def f_lte(x, y): return '1' if int(x) <= int(y) else ''
def f_gt(x, y): return '1' if int(x) > int(y) else ''
def f_gte(x, y): return '1' if int(x) >= int(y) else ''


# Boolean

def lazy_or(x, y): return str(x() or y())
def lazy_and(x, y): return str(x() and y())
def f_not(x): return '' if x else '1'


# Context operations

def context_get(ctx, name, default=''):
    if name in ctx and ctx[name]:
        return ctx[name]
    else:
        return default

def context_set(ctx, name, value):
    ctx[name] = value
    return ''

def context_unset(ctx, name):
    if name in ctx:
        del ctx[name]
    return ''


# Other

def lazy_noop(*args): return ''

def context_time(ctx, seconds):
    if seconds == '':
        return ''
    t = int(seconds)
    sign = ''
    if t < 0:
        sign = '-'
        t = -t
    t, s = divmod(t, 60)
    t, m = divmod(t, 60)
    d, h = divmod(t, 24)

    if d != 0:
        return ctx['__timefmt__'][0].format(sign, d, h, m, s)
    elif h != 0:
        return ctx['__timefmt__'][1].format(sign, h, m, s)
    else:
        return ctx['__timefmt__'][2].format(sign, m, s)





import mpd
from PySide.QtCore import *

from .connection import *
from .lists import *


class Server(ProperConnection, QObject):

    def __init__(self, main):
        super().__init__(main)

        State.create(self, 'volume', 50, 'setvol')
        State.create(self, 'current_song', {}, not_callable)
        BoolState.create(self, 'repeat')
        BoolState.create(self, 'shuffle', False, 'random')
        BoolState.create(self, 'single')
        BoolState.create(self, 'consume')
        BoolState.create(self, 'updating_db', False, not_callable)
        TimeTupleState.create(self, 'times', (0, 0), self._seek)

        self.queue = Queue(self)
        self.database = Database(self)
        self.playlists = Playlists(self)
        self.search = Search(self)

    def _seek(self, time):
        if self.state.value != 'stop':
            self.conn.seek(self.queue.current_pos, time)

    def update_state(self):
        s, c = {'state': 'disconnect'}, {}
        if self.state.value != 'disconnect':
            try:
                self.conn.command_list_ok_begin()
                self.conn.status()
                self.conn.currentsong()
                s, c = self.conn.command_list_end()
            except (mpd.MPDError, OSError) as e:
                self.handle_error(e)

        self.state.update(s['state'])
        self.times.update(s.get('time', ':').split(':'))
        self.volume.update(s.get('volume'))
        self.repeat.update(s.get('repeat'))
        self.shuffle.update(s.get('random'))
        self.current_song.update(c)
        self.single.update(s.get('single'))
        self.consume.update(s.get('consume'))
        self.updating_db.update('updating_db' in s)

        self.queue.current.update(s.get('playlist'))
        self.queue._current_pos.update(s.get('song'))

    @mpd_cmd(ignore_conn=True)
    def connect_mpd(self, host, port, password=None):
        if self.state.value != 'disconnect':
            self.disconnect_mpd()
        self.conn.connect(host, int(port))
        self.state.update('connect')
        if password:
            self.conn.password(password)

    @mpd_cmd
    def disconnect_mpd(self):
        try:
            try:
                self.conn.close()
            finally:
                self.conn.disconnect()
        finally:
            self.conn._reset()
            self.state.update('disconnect')

    play = simple_mpd('play')
    pause = simple_mpd('pause', 1)
    stop = simple_mpd('stop')
    previous = simple_mpd('previous')
    next = simple_mpd('next')
    clear = simple_mpd('clear')
    randomize_queue = simple_mpd('shuffle')

    @mpd_cmd
    def updatedb(self):
        if not self.updating_db.value:
            self.conn.update()

    @mpd_cmd(fallback={})
    def statistics(self):
        return self.conn.stats()





import pickle
from abc import ABCMeta, abstractmethod

import mpd
from PySide.QtCore import *

from .connection import *


class QABCMeta(ABCMeta, QObject.__class__):
    """A little hack to allow the use of ABCMeta."""


class SongList(RelayingConnection, QAbstractTableModel, metaclass=QABCMeta):

    current_pos = -1

    def __init__(self, parent, current, state_class=State):
        super().__init__(parent)
        self.setSupportedDragActions(Qt.MoveAction | Qt.CopyAction)
        self.items = []

        state_class.create(self, 'current', current, not_callable)
        self.current.changed.connect(self._update)
        self.state.changed2.connect(self._reset)

        self.total_length = 0

    def __len__(self):
        return len(self.items)

    def __getitem__(self, index):
        if type(index) == slice or 0 <= index < len(self):
            return self.items[index]
        return {}

    def __iter__(self):
        return iter(self.items)

    def refresh(self):
        self.current.update(self.current.value, force=True)

    def refresh_format(self):
        self.beginResetModel()
        self.endResetModel()

    def details(self, positions):
        if len(positions) == 1 and 'file' in self[positions[0]]:
            return self[positions[0]]

    @abstractmethod
    def ls(self, identifier):
        return []

    def can_add_to_queue(self, positions):
        return False

    def can_remove(self, positions):
        return False

    def can_rename(self, positions):
        return False

    def can_set_priority(self, positions, prio):
        return False

    @abstractmethod
    def item_chosen(self, pos):
        """i.e. double-clicked or Return pressed."""

    def sort_key(self, item):
        return (item.get('file', ''), item.get('directory', ''), item.get('playlist', ''))

    def _reset(self, newstate, oldstate):
        if newstate == 'disconnect' or oldstate == 'disconnect':
            self.current.update(None, force=True)

    def _update(self, new_current):
        self.main.busy()
        self.total_length = 0
        self.beginResetModel()
        self.items = []
        if self.state.value != 'disconnect':
            self.items = self.ls(new_current)
        self.endResetModel()
        for i in self.items:
            if 'time' in i:
                self.total_length += int(i['time'])
        self.main.not_busy()

    def columnCount(self, parent=None):
        return self.main.fmt.playlist_column_count

    def rowCount(self, parent=None):
        return len(self)

    def data(self, index, role=Qt.DisplayRole):
        r, c = index.row(), index.column()
        if 0 <= r < len(self):
            if role == Qt.DisplayRole:
                return self.main.fmt.playlist_item(
                        self[r], c, r == self.current_pos)
            elif role == Qt.ToolTipRole:
                return self.main.fmt.playlist_tooltip(
                        self[r], c, r == self.current_pos)
            elif role == Qt.DecorationRole:
                return self.main.fmt.playlist_icon(
                        self[r], c, r == self.current_pos)
            elif role == Qt.BackgroundRole:
                return self.main.fmt.playlist_background(
                        self[r], c, r == self.current_pos)
        return None

    def flags(self, index):
        f = Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsDragEnabled
        if index.column() == 0 and self.can_rename([index.row()]):
            f |= Qt.ItemIsEditable
        return f

    def mimeData(self, indexes):
        data = {'source': id(self), 'items': []}
        for i in indexes:
            if i.column() == 0:
                d = {'pos': str(i.row())}
                for k in ('id', 'playlist', 'directory', 'file'):
                    if k in self[i.row()]:
                        d[k] = self[i.row()][k]
                data['items'].append(d)
        data['items'].sort(key=lambda x: int(x['pos']))
        md = QMimeData()
        if len(data['items']) > 0:
            md.setData(self.MIMETYPE, pickle.dumps(data))
        return md

    def mimeTypes(self):
        return [self.MIMETYPE]

    def supportedDropActions(self):
        return None

    MIMETYPE = 'application/x-qygmy-playlistitems'


class WritableMixin(metaclass=ABCMeta):

    @mpd_cmd(fallback=False)
    def add(self, items, pos=None, extra=None):
        a = []
        for i in sorted(items, key=self.sort_key):
            if 'file' in i or self.can_add_directly(i, pos):
                a.append(i)
            elif 'directory' in i:
                a.extend(self.parent.database.ls(i['directory'], recursive=True))
            elif 'playlist' in i:
                a.extend(self.parent.playlists.ls(i['playlist']))
        if len(a) == 0:
            return False
        if pos is not None:
            a = reversed(a)
        with self.mpd_cmdlist:
            for i in a:
                self.add_one(i, pos, extra)
        self.refresh()
        return True

    @abstractmethod
    def can_add_directly(self, item, pos):
        return True

    @abstractmethod
    def add_one(self, item, pos, extra):
        pass

    def can_remove(self, positions):
        return len(positions) > 0

    @mpd_cmd
    def remove(self, positions):
        if self.can_remove(positions):
            items = []
            for pos in positions:
                i = self[pos].copy()
                i['pos'] = str(pos)
                items.append(i)
            with self.mpd_cmdlist:
                for i in reversed(sorted(items, key=lambda x: int(x['pos']))):
                    self.remove_one(i)
            self.refresh()

    @abstractmethod
    def remove_one(self, item):
        pass

    @mpd_cmd
    def move(self, items, pos):
        if len(items) == 0:
            return
        items.sort(key=lambda x: int(x['pos']))
        a = [i for i in items if int(i['pos']) <= pos]
        b = (i for i in items if int(i['pos']) > pos)
        apos = pos
        bpos = pos + (0 if len(a) == 0 else 1)
        with self.mpd_cmdlist:
            for i in reversed(a):
                if apos != i['pos']:
                    self.move_one(i, apos)
                apos -= 1
            for i in b:
                if bpos != i['pos']:
                    self.move_one(i, bpos)
                bpos += 1
        self.refresh()

    @abstractmethod
    def move_one(self, item, pos):
        pass

    def dropMimeData(self, data, action, row, column, parent):
        # Ignoring the action argument: move when src list is same as dst list,
        # copy otherwise.
        if not data.hasFormat(self.MIMETYPE):
            return False
        if row < 0:
            row = parent.row()
        if row < 0:
            row = None
        d = pickle.loads(data.data(self.MIMETYPE).data())
        if d['source'] == id(self):
            if row is None:
                return False
            self.move(d['items'], row)
        else:
            self.add(d['items'], row)
        return False

    def flags(self, index):
        return super().flags(index) | Qt.ItemIsDropEnabled

    def supportedDropActions(self):
        return Qt.MoveAction | Qt.CopyAction


class Queue(WritableMixin, SongList):

    def __init__(self, parent):
        super().__init__(parent, -1)
        State.create(self, '_current_pos', -1, 'play')
        self._current_pos.changed2.connect(self._update_current_pos)

    @property
    def current_pos(self):
        return self._current_pos.value

    def refresh(self):
        self.update_state()
        self.refresh_format()

    @mpd_cmd(fallback=[])
    def ls(self, _=None):
        return self.conn.playlistinfo()

    @mpd_cmd(fallback=False)
    def add(self, items, pos=None, play=False, replace=False):
        if replace:
            self.conn.clear()
        r = super().add(items, pos)
        if play and r:
            if replace:
                playpos = 0
            else:
                playpos = len(self) - 1 if pos is None else pos
            self.conn.play(playpos)
        return r

    def can_add_directly(self, item, pos):
        return 'file' in item or ('playlist' in item and pos is None)

    def add_one(self, item, pos, _):
        if 'playlist' in item and pos is None:
            self.conn.load(item['playlist'])
        elif 'file' in item:
            if pos is None:
                self.conn.add(item['file'])
            else:
                self.conn.addid(item['file'], pos)

    def remove_one(self, item):
        self.conn.deleteid(item['id'])

    def move_one(self, item, pos):
        self.conn.moveid(item['id'], pos)

    def can_set_priority(self, positions, prio):
        return any((self[i].get('prio', '0') == '0') != (int(prio) == 0) for i in positions)

    @mpd_cmd
    def set_priority(self, positions, prio):
        if self.can_set_priority(positions, prio):
            ids = [self[i]['id'] for i in positions]
            batch_size = self.MAX_MPD_ARGUMENTS - 1
            with self.mpd_cmdlist:
                for k in range((len(ids) - 1) // batch_size + 1):
                    b = ids[k * batch_size:(k+1) * batch_size]
                    self.conn.prioid(prio, *b)

    @mpd_cmd
    def reverse(self):
        with mpd_cmdlist:
            n = len(self)
            for i in range(n//2 + 1):
                self.conn.swap(i, n-i-1)

    def item_chosen(self, pos):
        self._current_pos.send(pos)

    def _update_current_pos(self, new, old):
        for i in {old, new}:
            if 0 <= i < len(self):
                index1 = self.index(i, 0)
                index2 = self.index(i, self.columnCount() - 1)
                self.dataChanged.emit(index1, index2)


class BrowserList(SongList):

    def can_add_to_queue(self, positions):
        return len(positions) > 0

    def add_to_queue(self, positions, play=False, replace=False):
        if self.can_add_to_queue(positions):
            self.parent.queue.add([self[i] for i in positions], play=play, replace=replace)

    def item_chosen(self, pos):
        self.add_to_queue([pos])

    def cd(self, new_current):
        self.current.update(new_current, force=True)


class Database(BrowserList):

    def __init__(self, parent):
        super().__init__(parent, '', PathState)

    def item_chosen(self, pos):
        if 'directory' in self[pos]:
            self.cd(self[pos]['directory'])
        else:
            super().item_chosen(pos)

    @mpd_cmd(fallback=[])
    def ls(self, path, recursive=False):
        l = self.conn.lsinfo(path)
        r = []
        for e in sorted(l, key=self.sort_key):
            if 'file' in e or (not recursive and 'directory' in e):
                r.append(e)
            elif 'directory' in e:
                r.extend(self.ls(e['directory'], recursive=True))
        return r


class Playlists(WritableMixin, BrowserList):

    def __init__(self, parent):
        super().__init__(parent, '')
        self.NEW_PLAYLIST_NAME = self.tr('New playlist')

    @mpd_cmd(fallback=[])
    def ls(self, name):
        if name != '':
            return self.conn.listplaylistinfo(name)
        return sorted(self.conn.listplaylists(), key=self.sort_key)

    def add(self, items, pos=None):
        extra = {}
        extra['last'] = len(self)
        if self.current.value != '':
            extra['name'] = self.current.value
        elif pos is not None:
            extra['name'] = self[pos]['playlist']
            pos = None
        else:
            playlists = {s['playlist'] for s in self}
            name = new = self.NEW_PLAYLIST_NAME
            if name in playlists:
                for i in range(2, 10000):
                    name = '{}_{}'.format(new, i)
                    if name not in playlists:
                        break
            extra['name'] = name
        return super().add(items, pos, extra=extra)

    def can_add_directly(self, item, pos):
        return 'file' in item

    def add_one(self, item, pos, extra):
        if 'file' not in item:
            return
        name = extra['name']
        if pos is not None:
            self.conn.playlistadd(name, item['file'])
            self.conn.playlistmove(name, extra['last'], pos)
            extra['last'] += 1
        else:
            self.conn.playlistadd(name, item['file'])

    def can_rename(self, positions):
        return len(positions) == 1 and 'playlist' in self[positions[0]]

    def item_chosen(self, pos):
        if 'playlist' in self[pos]:
            self.cd(self[pos]['playlist'])
        else:
            super().item_chosen(pos)

    def remove_one(self, item):
        if 'playlist' in item:
            self.conn.rm(item['playlist'])
        else:
            self.conn.playlistdelete(self.current.value, item['pos'])

    def move_one(self, item, pos):
        if self.current.value != '':
            self.conn.playlistmove(self.current.value, item['pos'], pos)

    @mpd_cmd
    def save_queue(self, name, replace=False):
        try:
            self.conn.save(name)
        except mpd.CommandError as e:
            if self.parse_exception(e)[0] != '56':
                raise
            if not replace:
                return False
            self.conn.rm(name)
            self.conn.save(name)
        if self.current.value in ('', name):
            self.refresh()
        return True

    def data(self, index, role=Qt.DisplayRole):
        r, c = index.row(), index.column()
        if role == Qt.EditRole and c == 0:
            return self[r].get('playlist', None)
        else:
            return super().data(index, role)

    @mpd_cmd(fallback=False)
    def setData(self, index, value, role=Qt.EditRole):
        r, c = index.row(), index.column()
        if role == Qt.EditRole and c == 0:
            name, new_name = self[r].get('playlist'), str(value)
            if name and name != new_name:
                self.conn.rename(name, new_name)
                self.refresh()
        return False


class Search(BrowserList):

    search_tags = [
        'any',
        'title',
        'artist',
        'album',
        'comment',
        'file',
    ]

    def __init__(self, parent):
        super().__init__(parent, (0, ''))

    @mpd_cmd(fallback=[])
    def ls(self, query):
        if query[1] == '':
            return []
        return sorted(self.conn.search(self.search_tags[query[0]], query[1]),
                key=self.sort_key)

